#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 2020, Jan Cervenka

import pandas as pd


def create_user_item_matrix(
    interactions: pd.DataFrame, user_column: str, item_column: str, feedback_column: str
) -> pd.DataFrame:
    """
    Creates `R` from `interactions` dataframe by pivoting it
    from long to wide format.

    Args:
        interactions: user-item feedback pairs
        user_column: name of the column containing user ids
        item_column: name of the column containing item ids
        feedback_column: name of the column containing feedbacks values

    Returns:
        R user-item matrix
    """

    # handles user-item duplicates by suming the feedback
    return interactions.pivot_table(
        index=user_column, columns=item_column, aggfunc="sum", values=feedback_column
    ).fillna(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 2020, Jan Cervenka

import pkg_resources

__module_name__ = "acf"

try:
    __version__ = pkg_resources.get_distribution(__module_name__).version
except Exception:
    __version__ = "unknown"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 2020, Jan Cervenka

import warnings

from typing import Tuple, Any

import numpy as np
import pandas as pd


def check_columns_in_dataframe(df: pd.DataFrame, columns: Tuple[str]) -> None:
    """
    Raises an exception if any column name in `columns` is not
    present in `df` dataframe.

    Args:
        df: dataframe to check
        columns: tuple of columns that required to be present
    """

    for col in columns:
        if col not in df.columns:
            raise ValueError(f"Column {col} is not in the dataframe.")


def check_feedback_column_numeric(
    interactions: pd.DataFrame, feedback_column: str
) -> None:
    """
    Raises an exception if `feedback_column` does not contain numeric values.

    Args:
        interactions: user-item interaction dataframe
        feedback_column: name of the column containing feedback values
    """

    if not pd.api.types.is_numeric_dtype(interactions[feedback_column]):
        raise ValueError(f'Column "{feedback_column}" must be numeric.')


def drop_warn_na(df: pd.DataFrame) -> pd.DataFrame:
    """
    Drops any rows with NA values and raises a warning.

    Args:
        df: dataframe to check

    Returns:
        df without the NA values
    """

    n_0 = len(df)
    df = df.dropna(how="any")

    if len(df) < n_0:
        warnings.warn(
            f"NA values found in the dataframe," f" {n_0 - len(df)} rows removed."
        )

    return df


def cast_numeric_greater_than_zero(
    value: Any, value_name: str, required_type: type
) -> None:
    """
    Checks that `value` is greater than zero and casts
    it to `required_type`.

    Raises an exception `value` not greater than zero.

    Args:
        value: numeric value to check
        value_name: name to be included in the error message
        required_type: target type of the value

    Returns:
        value as required type
    """

    if not isinstance(value, required_type):
        value = required_type(value)

    if value <= 0:
        raise ValueError(f"Value {value_name} must be greater than zero.")

    return value


def get_index_position(index: pd.Index, index_value: Any) -> np.int64:
    """
    Finds position of `index_value` in `index` array.
    This functions is used to find row number of given
    `user_id` or `item_id`.

    Raises an exception if `index` is not unique or the
    value is not found.

    Args:
        index: array of indeces
        index_value: value to find

    Returns:
        position of index_value as an integer
    """

    pos = np.where(index == index_value)[0]
    if len(pos) > 1:
        raise ValueError("Index is not unique.")

    if len(pos) == 0:
        raise ValueError(f"index_value = {index_value} not found in the index.")

    return pos[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 2020, Jan Cervenka

import unittest
from unittest import mock

import pandas as pd

from ..core.metrics import mean_rank


class MeanRankTest(unittest.TestCase):
    """
    Tests `metrics.mean_rank`.
    """

    def test_mean_rank_correct_value(self) -> None:
        """
        Tests that correct mean rank value is computed based
        on `interactions` and `X`, `Y` factors.
        """

        engine = mock.Mock()

        engine.user_factors = pd.DataFrame(
            {0: [0.5, 0.1, 0.9], 1: [0.1, 0.2, 0.5]}, index=["u1", "u2", "u3"]
        )

        engine.item_factors = pd.DataFrame(
            {0: [0.1, 0.2, 0.3, 0.3], 1: [0.5, 0.9, 0.9, 0.7]},
            index=["i1", "i2", "i3", "i4"],
        )

        test_interactions = pd.DataFrame(
            {
                "user_id": ["u1", "u1", "u1", "u2", "u2", "u3", "u3", "u3", "u3"],
                "item_id": ["i1", "i2", "i4", "i2", "i3", "i1", "i2", "i3", "i4"],
                "feedback": [4, 2, 1, 5, 10, 9, 8, 12, 8],
            }
        )

        # R =
        # 4 2  0 1
        # 0 5 10 0
        # 9 8 12 8

        # X * Y^T =
        # 1.00 0.75 0.25 0.50
        # 0.11 0.20 0.21 0.17
        # ą0.34 0.63 0.72 0.62

        # X * Y^T ranks
        # 1.00 0.75 0.25 0.50
        # 1.00 0.50 0.25 0.75
        # 1.00 0.50 0.25 0.75

        result = mean_rank(
            interactions=test_interactions,
            user_column="user_id",
            item_column="item_id",
            feedback_column="feedback",
            engine=engine,
        )

        expected = 33 / 59
        self.assertEqual(result, expected)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 2020, Jan Cervenka

import multiprocessing as mp
import unittest

from typing import Iterable

import numpy as np
import pandas as pd

from ..core.computation import Engine


class EngineTest(unittest.TestCase):
    """
    Tests `computation.Engine` class.
    """

    def setUp(self) -> None:
        """
        Sets up the tests.
        """

        self._engine = Engine(random_state=0)

    def test_initialize_factors(self) -> None:
        """
        Tests that `Engine._initialize_factors` produce factor
        matrices `X` and `Y` with random values in correct shape.
        """

        test_m, test_n = (15, 25)

        X_expected, Y_expected = self._engine._initialize_factors(test_m, test_n)

        for expected, size in ((X_expected, test_m), (Y_expected, test_n)):
            self.assertTupleEqual((size, self._engine._n_factors), expected.shape)

    def test_feedback_1d_generator(self) -> None:
        """
        Tests that `Engine._feedback_1d_generator` produces correct
        row/column arrays from `R` based on the `axis` argument.
        """

        def assert_collection_of_numpy_equal(
            x: Iterable[np.array], y: Iterable[np.array]
        ) -> None:
            """
            Asserts that collections `x` and `y` containing
            numpy arrays are identical.
            """

            x, y = list(x), list(y)
            assert len(x) == len(x)

            for array_x, array_y in zip(x, y):
                np.testing.assert_array_equal(array_x, array_y)

        test_R = pd.DataFrame({"A": [1, 3], "B": [6, 7]})

        tests = (
            ([np.array([1, 3]), np.array([6, 7])], 0),
            ([np.array([1, 6]), np.array([3, 7])], 1),
        )

        for expected, axis in tests:
            result = list(self._engine._feedback_1d_generator(test_R, axis))
            assert_collection_of_numpy_equal(result, expected)

    def test_compute_factors_1d(self) -> None:
        """
        Tests that `Engine._compute_factors` computes correct
        factor row for one user/item.
        """
        np.random.seed(0)

        test_feedback_1d = np.array([5, 4, 0, 0, 0, 4, 0, 0, 0, 0])
        test_other_factors = np.random.rand(10, 5)
        test_other_factors_small = np.dot(test_other_factors.T, test_other_factors)

        result = self._engine._compute_factors_1d(
            feedback_1d=test_feedback_1d,
            other_factors=test_other_factors,
            other_factors_small=test_other_factors_small,
            reg_lambda=1,
            alpha=40,
        )

        expected = np.array([0.2940538, 0.52162197, 0.82403064, -0.2290155, 0.17128187])

        np.testing.assert_array_almost_equal(result, expected, decimal=7)

    def test_compute_factors(self) -> None:
        """
        Tests that `Engine._compute_factors` computes
        correct factor rows for every user/item.
        """
        np.random.seed(0)

        R_test = pd.DataFrame(
            [
                [3, 2, 0, 0, 0, 0, 0, 0, 1, 2],
                [0, 2, 3, 0, 0, 0, 0, 2, 3, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 2],
                [0, 1, 0, 2, 2, 0, 0, 0, 2, 2],
                [1, 0, 0, 0, 2, 0, 3, 0, 0, 0],
                [1, 0, 0, 0, 2, 3, 0, 3, 0, 0],
                [0, 0, 3, 0, 2, 0, 0, 0, 0, 0],
                [0, 0, 1, 3, 3, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            ]
        )

        test_n = R_test.shape[-1]
        # axis=1 test for user factors, other_factors matrix is for items
        test_axis = 1
        test_other_factors = np.random.rand(test_n, 3)

        pool = mp.Pool(1)
        try:
            result = self._engine._compute_factors(
                pool=pool, R=R_test, other_factors=test_other_factors, axis=test_axis
            )
        finally:
            pool.close()

        # shape (9, 3): 3 factors for 9 users
        expected = np.array(
            [
                [0.35764017, 1.03013196, 0.27789359],
                [1.21552231, 1.36885415, -0.71711505],
                [1.5414733, -0.52604296, -0.47427519],
                [0.24321584, 1.0214088, 0.40020838],
                [0.04003445, 1.06238047, 0.06082416],
                [-0.51951887, 1.15240956, 1.12119426],
                [-0.79891536, 1.55973535, -0.05766712],
                [-1.09061448, 1.76283214, -0.04719709],
                [0.27658886, 0.94346013, -0.90276183],
            ]
        )

        np.testing.assert_array_almost_equal(result, expected, decimal=7)

    @staticmethod
    def test_predict() -> None:
        """
        Tests that `Engine.predict` computes correct
        recommendation for given `user`.
        """

        engine = Engine()
        engine._X = np.array([[0.50, 0.90, 0.30], [0.05, 0.94, 0.81]])

        engine._Y = np.array(
            [[0.20, 0.12, 0.80], [0.50, 0.97, 0.03], [0.75, 0.02, 0.15]]
        )

        # user 20 likes factors 1 and 2
        # item 4 really belongs to genre 1 and item 3 belongs to genre 2
        # these two items will be recommended
        #
        # user 20 doesnt care about factor 0 where item 5 belongs
        # small recommendation for item 5

        engine._user_index = pd.Index([10, 20], name="user_id")
        engine._item_index = pd.Index([3, 4, 5], name="item_id")

        result = engine.predict(user=20)
        result_top_2 = engine.predict(user=20, top_n=2)

        expected = pd.Series(
            [0.7708, 0.9611, 0.1778], index=engine._item_index, name=20
        )
        # calling .nlargest on series will sort the values (descending)
        expected_top_2 = expected[[True, True, False]].sort_values(ascending=False)

        pd.testing.assert_series_equal(result, expected)
        pd.testing.assert_series_equal(result_top_2, expected_top_2)

    def test_get_loss(self) -> None:
        """
        Tests that `Engine._get_loss` computes correct least squares loss.
        """

        R_test = np.array([[1, 0, 0, 4], [2, 1, 0, 0], [0, 0, 0, 1]])

        X_test = np.array([[0.3, 0.1], [0.9, 0.4], [0.1, 0.5]])

        Y_test = np.array([[0.4, 0.1], [0.9, 0.5], [0.9, 0.9], [0.6, 0.5]])

        result = self._engine._get_loss(X_test, Y_test, R_test)
        expected = 177.73779
        self.assertAlmostEqual(result, expected, places=4)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 2020, Jan Cervenka

from .core.computation import Engine
from .core import metrics

from .version import __version__

__doc__ = """
Lightweight recommender engine for implicit feedback datasets.

The package implements a collaborative filtering algorithm as described
in "Collaborative Filtering for Implicit Feedback Datasets" paper by
Yifan Hu, Yehuda Koren, Chris Volinsky (https://doi.org/10.1109/ICDM.2008.22).
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 2020, Jan Cervenka

import unittest

import numpy as np
import pandas as pd

from ..core.preprocessing import create_user_item_matrix


class CreateUserItemMatrixTest(unittest.TestCase):
    """
    Tests `preprocessing.create_user_item_matrix`.
    """

    @staticmethod
    def test_create_user_item_matrix() -> None:
        """
        Tests that the created user-item matrix is correct and
        that duplicate user-item pairs are correctly aggregated.
        """

        test_interactions = pd.DataFrame(
            {
                "id_item": ["A", "B", "C", "A", "B", "A", "D", "D"],
                "id_user": [1, 1, 1, 2, 2, 3, 4, 4],
                "feedback": [6, 2, 1, 5, 4, 2, 1, 2],
            }
        )

        result = create_user_item_matrix(
            test_interactions, "id_user", "id_item", "feedback"
        )

        expected_matrix = np.array(
            [[6, 2, 1, 0], [5, 4, 0, 0], [2, 0, 0, 0], [0, 0, 0, 3]], dtype="float64"
        )

        expected = pd.DataFrame(
            expected_matrix,
            index=pd.Index([1, 2, 3, 4], name="id_user"),
            columns=pd.Index(["A", "B", "C", "D"], name="id_item"),
        )

        pd.testing.assert_frame_equal(result, expected)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 2020, Jan Cervenka

import logging
import functools
import multiprocessing as mp

from typing import Any, Callable, Iterable, List, Optional, Tuple

import numpy as np
import pandas as pd

from .preprocessing import create_user_item_matrix
from .utils import (
    cast_numeric_greater_than_zero,
    check_columns_in_dataframe,
    check_feedback_column_numeric,
    get_index_position,
    drop_warn_na,
)

DEFAULT_REG_LAMBDA = 0.1
DEFAULT_ALPHA = 40
DEFAULT_N_FACTORS = 10
DEFAULT_N_ITER = 20
DEFAULT_N_JOBS = 1

# TODO: sparse R dataframe/matrix?
# TODO: X/Y init - which distributions?
# TODO: _compute_factors_1d, _get_least_square_sum to algebra.py module?


class Engine:
    """
    Collaborative filtering algorithm based on matrix factorization
    by alternating least squares. Designed for implicit feedback datasets.

    The class exposes two public methods, `fit` to train the model and
    `predict` to produce the recommendations.

    Example:
    ```
    import acf
    import pandas as pd

    # assuming the data are in the following format:
    # | user_id | item_id | feedback |
    # |---------|---------|----------|
    # | 2491    | 129     | 2        |

    interactions = pd.read_csv('interactions.csv')

    engine = acf.Engine(reg_lambda=1, alpha=35, n_factors=2)
    engine.fit(interactions,
               user_column='user_id',
               item_column='item_id',
               feedback_column='feedback',
               n_iter=20,
               n_jobs=4)

    # get the best 20 recommendation for given user
    prediction = engine.predict(user=2491, top_n=20)

    # to print training loss value at every iteration
    print(engine.loss)
    ```
    """

    def __init__(
        self,
        reg_lambda: float = DEFAULT_REG_LAMBDA,
        alpha: float = DEFAULT_ALPHA,
        n_factors: int = DEFAULT_N_FACTORS,
        random_state: Optional[int] = None,
    ):
        """
        Initializes the class with model hyperparameters
        and `random_state`.

        Args:
            reg_lambda: regularization parameter
            alpha: confidence control parameter
            n_factors: number of latent factors
            random_state: RNG seed
        """

        self._reg_lambda = cast_numeric_greater_than_zero(
            reg_lambda, "reg_lambda", float
        )
        self._alpha = cast_numeric_greater_than_zero(alpha, "alpha", float)
        self._n_factors = cast_numeric_greater_than_zero(n_factors, "n_factors", int)

        if random_state:
            self._rng = np.random.RandomState(random_state)
        else:
            self._rng = np.random.RandomState()

        self._X = self._Y = None
        self._user_index = self._item_index = None
        self._loss = []

    @staticmethod
    def _compute_factors_1d(
        feedback_1d: np.array,
        other_factors: np.array,
        other_factors_small: np.array,
        reg_lambda: float,
        alpha: float,
    ) -> np.array:
        """
        Computes a 1-dimensional factor array for either one user
        or one item.

        When computing user factors, `feedback_1` is an array of interactions
        `r_u` between user `u` and every item, `other_factors` is an item
        factor matrix `Y` and `other_factors_small` is `Y ^ T * Y`.

        When computing the user factors `x_u` for user `u`, the function
        solves a linear system

        ```
        (Y^T * Y + Y^T * (C^u - I) * Y + λ * I) * x_u = Y^T * C^U * p_u
        ```

        where `Y` is an item factor matrix, `C^u` is a diagonal matrix
        containing feedback `r_u` of user `u` for each item `i` (the
        feedback is transformed to `c_ii = 1 + alpha * r_ui`), `p_u`
        is a binary vector of preferences of user `u` for each item,
        and `λ` is regularization lambda.

        The computation is symmetric for item factors.

        Args:
            feedback_1d: one R user-item matrix row or column
            other_factors: either X or Y factor matrix
            other_factors_small: either X^T * X or Y^T * Y matrix
            :reg_lambda: regularization parameter
            alpha: confidence control parameter

        Returns:
            1-dimensional factor array for one user/item
        """
        p_ = (feedback_1d > 0).astype("uint8")
        C_ = np.diag(1 + alpha * feedback_1d)

        size, f = other_factors.shape

        M = other_factors_small + np.linalg.multi_dot(
            [other_factors.T, C_ - np.eye(size), other_factors]
        )

        return np.linalg.solve(
            M + reg_lambda * np.eye(f), np.linalg.multi_dot([other_factors.T, C_, p_])
        )

    def _create_worker(
        self, other_factors: np.array, other_factors_small: np.array
    ) -> Callable:
        """
        Creates a callable worker from `Engine._compute_factors_1d` for
        application over `R` rows/columns in multiprocessing map.

        The worker takes only one argument (`R` column/row), other
        parameters are set in this function using `functools.partial`
        and are identical for all given rows/columns.

        Args:
            other_factors: either X or Y factor matrix
            other_factors_small: either X^T * X or Y^T * Y matrix

        Returns:
            callable worker
        """

        return functools.partial(
            self._compute_factors_1d,
            other_factors=other_factors,
            other_factors_small=other_factors_small,
            reg_lambda=self._reg_lambda,
            alpha=self._alpha,
        )

    @staticmethod
    def _feedback_1d_generator(R: pd.DataFrame, axis: int) -> Iterable[np.array]:
        """
        Creates a generator from `R` rows or columns.
        Each item in the generator is a numpy array.

        Args:
            R: user-item feedback matrix
            axis: axis=1 for rows, axis=0 for columns

        Returns:
            R row/column generator
        """

        for _, r in R.iterrows() if axis else R.iteritems():
            yield r.values

    def _compute_factors(
        self, pool, R: pd.DataFrame, other_factors: np.array, axis: int
    ) -> np.array:
        """
        Applies `Engine._compute_factors_1d` over `R` rows/columns
        to compute factors for all user/items.

        Args:
            pool: multiprocessing pool
            R: user-item feedback matrix
            other_factors: either X or Y factor matrix
            axis: axis=1 for rows, axis=0 for columns

        Returns:
            either X or Y factor matrix
        """
        # TODO: axis can be infered from other_factors shape

        other_factors_small = np.dot(other_factors.T, other_factors)
        worker = self._create_worker(other_factors, other_factors_small)

        all_feedback_1d = self._feedback_1d_generator(R, axis)

        return np.vstack(pool.map(worker, all_feedback_1d))

    def _initialize_factors(self, m: int, n: int) -> Tuple[np.array]:
        """
        Initialzes `X`, `Y` matrices with random values.

        Args:
            m: number of users
            n: number of items

        Returns:
            tuple of X, Y factor matrices
        """

        return (self._rng.rand(m, self._n_factors), self._rng.rand(n, self._n_factors))

    def _get_loss(self, X: np.array, Y: np.array, R: np.array) -> np.float:
        """
        Computes least squares sum for given `X`, `Y`, `R`.

        The loss is defined as `loss := Σ_ui (C * (P - X * Y^T)^2)`
        where `C` is confidence matrix, `P` is preference matrix,
        and `X`, `Y` are factor matrices.

        Args:
            X: user factors matrix
            Y: item factors matrix
            R: user-item feedback matrix

        Returns:
            loss value
        """

        P = (R > 0).astype("uint8")
        C = 1 + self._alpha * R

        return np.sum(C * (P - np.dot(X, Y.T)) ** 2)

    @staticmethod
    def _log_iteration(iteration: int, loss: np.float) -> None:
        """
        Logs least squares `loss` for given `iteration`.

        Args:
            iteration: training iterataion
            loss: value to log
        """

        logging.info(f"Iteration {iteration:04d} Loss Σ = {loss:.5f}")

    def _run_iterations(self, pool, R: pd.DataFrame, n_iter: int) -> Tuple[np.array]:
        """
        Initializes `X`, `Y` matrices and runs
        alternating least squares iterations.

        Args:
            pool: multiprocessing pool
            R: user-item feedback matrix
            n_iter: number of iterations

        Returns:
            tuple of X, Y factor matrices
        """

        X, Y = self._initialize_factors(*R.shape)

        for i in range(n_iter):
            X = self._compute_factors(pool, R, Y, axis=1)  # rows
            Y = self._compute_factors(pool, R, X, axis=0)  # columns
            self._loss.append(self._get_loss(X, Y, R.values))
            self._log_iteration(i, self._loss[-1])

        return X, Y

    def fit(
        self,
        interactions: pd.DataFrame,
        user_column: str,
        item_column: str,
        feedback_column: str,
        n_iter: int = DEFAULT_N_ITER,
        n_jobs: int = DEFAULT_N_JOBS,
    ) -> None:
        """
        Fits the model by factorizing `interactions` into latent factors.

        Dataframe `interactions` contains implicit feedbacks values for
        user-item pairs in three columns `user_column`, `item_column`,
        `feedback_column`.

        ```
        | user_column | item_column | feedback_column    |
        |-------------|-------------|--------------------|
        | user_id_0   | item_id_0   | 0_1_feedback_value |
        ```

        Args:
            interactions: user-item feedback pairs
            user_column: name of the column containing user ids
            item_column: name of the column containing item ids
            feedback_column: name of the column containing feedbacks values
            n_iter: number of alternating least squares iterations
            n_jobs: number of multiprocessing jobs
        """

        check_columns_in_dataframe(
            interactions, (user_column, item_column, feedback_column)
        )
        check_feedback_column_numeric(interactions, feedback_column)

        interactions = drop_warn_na(
            interactions[[user_column, item_column, feedback_column]]
        )

        n_iter = cast_numeric_greater_than_zero(n_iter, "n_iter", int)

        R = create_user_item_matrix(
            interactions=interactions,
            user_column=user_column,
            item_column=item_column,
            feedback_column=feedback_column,
        )

        self._user_index, self._item_index = R.index, R.columns

        pool = mp.Pool(n_jobs)

        try:
            self._X, self._Y = self._run_iterations(pool, R, n_iter)

        finally:
            pool.close()

    def predict(self, user: Any, top_n: Optional[int] = None) -> pd.Series:
        """
        Computes recommendations for given `user`.

        Args:
            user: target of the recommendations
            top_n: if not `None`, selects only the best n items

        Returns:
            series with predicted score for each item
        """

        row = get_index_position(self._user_index, user)

        prediction = pd.Series(
            np.dot(self._Y, self._X[row, :]), index=self._item_index, name=user
        )

        if top_n:
            return prediction.nlargest(top_n)
        else:
            return prediction

    @property
    def user_factors(self) -> pd.DataFrame:
        """
        User factors property.

        Returns:
            user factors as a dataframe
        """

        return pd.DataFrame(self._X, index=self._user_index)

    @property
    def item_factors(self) -> pd.DataFrame:
        """
        Item factors property.

        Returns:
            item factors as a dataframe
        """

        return pd.DataFrame(self._Y, index=self._item_index)

    @property
    def loss(self) -> List[np.float]:
        """
        Training loss proprety.

        Returns:
            training loss by iteration
        """

        return self._loss



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 2020, Jan Cervenka

import setuptools

VERSION = "0.2.2"


def run_setup():
    """
    Runs the package setup.
    """

    setup_params = {
        "name": "acf",
        "version": VERSION,
        "description": "Lightweight recommender engine",
        "author": "Jan Cervenka",
        "author_email": "jan.cervenka@yahoo.com",
        "packages": ["acf", "acf.core", "acf.tests"],
        "python_requires": ">=3.7",
        "install_requires": ["pandas>=1.0", "numpy>=1.16"],
    }
    setuptools.setup(**setup_params)


if __name__ == "__main__":
    run_setup()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 2020, Jan Cervenka

import numpy as np
import pandas as pd

from . import computation

from .preprocessing import create_user_item_matrix


def mean_rank(
    interactions: pd.DataFrame,
    user_column: str,
    item_column: str,
    feedback_column: str,
    engine: computation.Engine,
) -> np.float:
    """
    Computes mean rank metric using real `interactions`
    as ground truth and predictions produced by `engine`.

    The metric is defined as

    ```
    mean_rank := Σ_ui r_ui * rank_ui / Σ_ui r_ui
    ```

    where `r_ui` is the feedback value between user `u` and item `i`
    and `rank_ui` is the recommendation rank of item `i` for user `u`.

    `rank_ui` is computed by inverse row-wise percentile ranking of
    values in `X * Y^T` prediction matrix. Value `r_ui = 0` means
    that item `i` is the first to be recommended for user `u`,
    `r_uj = 1` is the last to be recommended.

    The metric is a mean rank value weighted by `R` feedbacks.

    Args:
        interactions: user-item feedback pairs
        user_column: name of the column containing user ids
        item_column: name of the column containing item ids
        feedback_column: name of the column containing feedbacks values
        engine: trained model

    Returns:
        computed metric
    """

    # TODO: column/NA checks

    R = create_user_item_matrix(
        interactions=interactions,
        user_column=user_column,
        item_column=item_column,
        feedback_column=feedback_column,
    )

    user_ids, item_ids = R.index.tolist(), R.columns.tolist()

    # prediction = X * Y ^ T
    # ranks = apply rank row wise on prediction
    ranks = (
        engine.user_factors.loc[user_ids, :]
        .dot(engine.item_factors.loc[item_ids, :].T)
        .rank(pct=True, ascending=False, axis=1)
    )

    # multiply ranks with R element-wise, make a sum and divide by R sum
    return R.mul(ranks).sum().sum() / R.sum().sum()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 2020, Jan Cervenka

import unittest

import numpy as np
import pandas as pd

from ..core import utils


class CheckColumnsInDataFrameTest(unittest.TestCase):
    """
    Tests `utils.check_columns_in_data_frame`.
    """

    def setUp(self) -> None:
        """
        Sets up the tests.
        """

        self._test_df = pd.DataFrame({"x": [1], "y": [1], "z": [1]})

    def test_all_columns_in_dataframe(self) -> None:
        """
        Tests that no exception is raised if a column is present.
        """

        tests = (tuple(), ("x", "y"), ("x",), ("x", "y", "z"))
        for columns in tests:
            utils.check_columns_in_dataframe(self._test_df, columns)

    def test_column_not_in_data_frame(self) -> None:
        """
        Tests that an exception is raised if a column is not present.
        """

        for columns in (("a"), ("x", "a")):
            with self.assertRaises(ValueError):
                utils.check_columns_in_dataframe(self._test_df, columns)


class CheckFeedbackColumnNumericTest(unittest.TestCase):
    """
    Tests `utils.check_feedback_column_numeric`.
    """

    def setUp(self) -> None:
        """
        Sets up the tests.
        """

        self._test_df = pd.DataFrame({"x": ["a", "b"], "y": [1, 3]})

    def test_feedback_column_is_numeric(self) -> None:
        """
        Tests that no exception is raised if the feedback column is numeric.
        """

        utils.check_feedback_column_numeric(self._test_df, "y")

    def test_feedback_column_not_numeric(self) -> None:
        """
        Tests that a exception is raised if the feedback column is not
        numeric.
        """

        with self.assertRaises(ValueError):
            utils.check_feedback_column_numeric(self._test_df, "x")


class DropWarnNaTest(unittest.TestCase):
    """
    Tests `utils.test_drop_warn_na`.
    """

    def test_na_present(self) -> None:
        """
        Tests that rows are dropped and warning is raised when NA
        value is present.
        """

        test_df = pd.DataFrame({"x": [1.0, 2.0, np.nan]})
        expected = pd.DataFrame({"x": [1.0, 2.0]})

        with self.assertWarns(UserWarning):
            result = utils.drop_warn_na(test_df)

        pd.testing.assert_frame_equal(result, expected)

    @staticmethod
    def test_na_not_present() -> None:
        """
        Tests that original dataframe is returned when no NA is present.
        """

        test_df = expected = pd.DataFrame({"x": [1, 2]})
        expected = pd.DataFrame({"x": [1, 2]})

        result = utils.drop_warn_na(test_df)
        pd.testing.assert_frame_equal(result, expected)


class CastNumericGreaterThanZeroTest(unittest.TestCase):
    """
    Tests `utils.test_drop_warn_na`.
    """

    def test_value_converted(self) -> None:
        """
        Tests that compatible `value` are correctly cast to `required_type`.
        """

        tests = ((1, "test", int, 1), (2.3, "test", float, 2.3), (1.2, "test", int, 1))

        for value, value_name, required_type, expected in tests:
            result = utils.cast_numeric_greater_than_zero(
                value, value_name, required_type
            )

            self.assertEqual(result, expected)

    def test_value_not_numeric(self) -> None:
        """
        Test that an exception is raised when a non-numeric `value`
        cannot be cast to `required_type`.
        """

        with self.assertRaises(ValueError):
            _ = utils.cast_numeric_greater_than_zero("test", "test", float)

    def test_value_not_greater_than_zero(self) -> None:
        """
        Test that an exception is raised when a numeric `value`
        is not greater than zero.
        """

        with self.assertRaises(ValueError):
            _ = utils.cast_numeric_greater_than_zero(-1, "test", float)


class GetIndexPositionTest(unittest.TestCase):
    """
    Tests `utils.get_index_position`.
    """

    def test_index_value_not_unique(self) -> None:
        """
        Tests that an exception is raised when
        `index_value` is not unique in `index`.
        """

        test_index = pd.Index([1, 3, 2, 2])
        test_index_value = 2

        with self.assertRaises(ValueError):
            _ = utils.get_index_position(test_index, test_index_value)

    def test_index_value_not_found(self) -> None:
        """
        Tests that an exception is raised when
        `index_value` is not present in `index`.
        """

        test_index = pd.Index([1, 3, 2])
        test_index_value = 4

        with self.assertRaises(ValueError):
            _ = utils.get_index_position(test_index, test_index_value)

    def test_index_value_found(self) -> None:
        """
        Tests that correct row positon is returned.
        """

        test_index = pd.Index([1, 3, 2, 4])
        test_index_value = 3

        result = utils.get_index_position(test_index, test_index_value)
        expected = 1

        self.assertEqual(result, expected)


if __name__ == "__main__":
    unittest.main()
